struct tab_simb *insere_nodo ();
void poe_valor (int v);
void incr_ref(struct tab_simb *p);
void poe_linha ();
void poe_token (int tk);
void poe_tipo (int t);
void desemp_proc(void);
void desaloca_tab_simb(void);
struct tab_simb *pesquisa_local(char *p);
struct tab_simb *pesquisa_global(char *p);
struct tab_simb *pesquisa_nivel(char *p);

#define TIPOVAR       0
#define TIPOPROC      1
#define TIPOCONST     2


typedef struct tab_simb {
        char *nome;       /* nome do ident, proc ou const */
		int lin,          /* linha de def. */
            nref,         /* numero de referencias ao ID */
            nivel,        /* nivel de aninhamento */
            desl,         /* deslocamento da variavel no procedimento */
            valor,        /* valor se const */
            token,        /* IF, DO, etc    */
            tipo;         /* tipo: 0 - variavel
                                   1 - procedimento
                                   2 - constante
                          */

	 struct tab_simb *prox;
};

char *proc_corr[500]; /* nome do procedimento sendo compilado, */
		      /* possibilita aninhamento maximo de 500 procedimentos */
int nivel=0; /* nivel atual */
int desl=0;  /* deslocamento atual na declaracao */
int decl=0;  /* se 1:area de declaracao, 0:c.c. */
int erros=0; /* numeros de erros na compilacao */
int numlin=0; /* numero da linha para depuracao */
char label[6], label1[6], /* rotulos para geracao de codigo */
     iniproc[500][6]; /* rotulo para fazer um jmp sempre para o inicio */
		      /* do codigo do procedimento quando ha proc. aninhados */
unsigned procnro=0; /* usado para indice */


struct tab_simb *corr, /* nodo corrente da tab. de simb. */
		*aux, *aux2; /* auxiliares */

void incr_ref(struct tab_simb *p)
{ /* incrementa o numero de referencias a um dado id */
  p->nref++;
}

void poe_linha ()
{ /* poe o numero da linha da declaracao do id */
  corr->lin = numlin;
}

void poe_valor (int v)
{  /* poe o valor decimal quando for uma constante */
  corr->valor = v;
}

void poe_tipo (int t)
{ /* poe o tipo : const, var, procedimento */
  corr->tipo = t;
}

void poe_token (int tk)
{ /* poe o token: CALL, IF, DO, etc */
  corr->token = tk;
}

void poe_desl(void)
{ /* poe o deslocamento da declaracao no corrente procedimento */
  corr->desl = desl;
  desl+=2;
}

void poe_nivel(void)
{  /* poe o nivel de aninhamento */
  corr->nivel = nivel;
}

struct tab_simb *insere_nodo(char *s)
{  /* insere um novo nodo na tab de simbolo com o nome s */
   static primeiro=1;

   struct tab_simb *aux = calloc (1,sizeof(struct tab_simb));
   if(primeiro){
     corr=aux;
     corr->prox=NULL;
     primeiro=0;
     strcpy (corr->nome = malloc (strlen(s)+1),strlwr(s));
     corr->nref=0;
     return corr;
   }
   if (aux == NULL){
     printf("Erro de alocacao de memoria para compilacao.\n");
     if (!primeiro) desaloca_tab_simb();
     exit(1);
   }else {
     aux->prox = corr;
     corr=aux;
     strcpy (corr->nome = malloc (strlen(s)+1),strlwr(s));
     corr->nref=0;
     return corr;
   }
}

struct tab_simb *pesquisa_global(char *p)
{ /* faz uma pesquisa em toda a tabela */
  struct tab_simb *aux;

  for (aux=corr; aux != NULL; aux=aux->prox)
      if (!strcmp (aux->nome,strlwr(p)))
         return (aux);
  return NULL;
}

struct tab_simb *pesquisa_local(char *p)
{ /* pesquisa so no procedimento atual */
  struct tab_simb *aux;

  for(aux=corr; aux != NULL && aux->tipo != 1; aux=aux->prox)
      if (!strcmp (aux->nome,strlwr(p)))
         return (aux);
  return NULL;
}

struct tab_simb *pesquisa_nivel(char *p)
{ /* pesquisa so no nivel atual */
  struct tab_simb *aux;

  for(aux=corr; aux->nivel==corr->nivel && aux; aux=aux->prox)
    if(!strcmp(aux->nome,strlwr(p)))
      return aux;
  return NULL;
}

void desemp_proc(void)
{ /* desaloca o ultimo procedimento compilado */
  struct tab_simb *aux, *aux2;

  aux=pesquisa_global(strlwr(proc_corr[nivel]));
  for(aux2=corr; aux2!=aux; aux2=corr){
    corr=corr->prox;
    free(aux2);
  }
  strupr(proc_corr[nivel]);
}

void desaloca_tab_simb(void)
{ /* desaloca toda a tabela de simbolos */
  struct tab_simb *aux;

  for (aux=corr; aux != NULL; aux=corr) {
      corr = corr->prox;
      free (aux);
  }
}



